(* 多相型関数をインスタンス化する *)

open Type
open Typing
open Id
open FlatSyntax

module M3 = MyMap.Make(struct
  type t = Type.t
  let compare = Pervasives.compare
end)

(* 型変数と代入される型の対応を作る *)
let make_tenv t1 t2 =
  let rec aux acc t1 t2 =
    match t1,t2 with
    | Tvar({ contents = None } as v), t ->
        if List.mem_assq v acc then acc else (v,t) :: acc
    | Tarray t1, Tarray t2 -> aux acc t1 t2
    | Ttuple t1s, Ttuple t2s -> List.fold_left2 aux acc t1s t2s
    | Tfun(t1s,t1), Tfun(t2s,t2) -> List.fold_left2 aux (aux acc t1 t2) t1s t2s
    | _ -> assert (t1 == t2); acc
  in
  aux [] t1 t2

(* 型変数に代入した新しい型を作る *)
let rec copy_t tenv t =
  match t with
  | Tvar({ contents = None } as v) ->
      (try List.assq v tenv with Not_found -> t)
  | Tarray t -> Tarray(copy_t tenv t)
  | Ttuple ts -> Ttuple(List.map (copy_t tenv) ts)
  | Tfun(ts,t) -> Tfun(List.map (copy_t tenv) ts, copy_t tenv t)
  | ForAll(vs,t) -> ForAll(vs, copy_t tenv t)
  | Tvar _ -> assert false
  | _ -> t

let copy_id tenv x = new_id_t x.name (copy_t tenv x.typ)

(* 型変数に代入した新しい式を作る *)
let rec copy env tenv e =
  let find x = match x with
  | Var y -> (try Var(M.find y env) with Not_found -> x)
  | _ -> x
  in
  match e with
  | Let(xs,e1,e2) ->
      let xs_ = List.map (copy_id tenv) xs in
      Let(xs_, copy env tenv e1, copy (M.add_list2 xs xs_ env) tenv e2)
  | LetRec(x,ys,e1,e2) ->
      let x_ = copy_id tenv x in
      let ys_ = List.map (copy_id tenv) ys in
      let env2 = M.add x x_ env in
      let env1 = M.add_list2 ys ys_ env2 in
      LetRec(x_, ys_, copy env1 tenv e1, copy env2 tenv e2)
  | If(c,x,y,e1,e2) ->
      If(c, find x, find y, copy env tenv e1, copy env tenv e2)
  | Op(x,ys) -> Op(find x, List.map find ys)
  | Tuple xs -> Tuple(List.map find xs)
  | V x -> V(find x)
  | ExtApp(x,ys) -> ExtApp(x, List.map find ys)

(* 多相型関数と型から、すでに生成された関数への写像 *)
let env = ref M.empty

let add x t y = env :=
  M.add x (M3.add t y (try M.find x !env with Not_found -> M3.empty)) !env
let find x t =
  try M3.find t (M.find x !env) with Not_found ->
    let y = new_id_t x.name t in
    add x t y; y

let hoge x t =
  match x.typ with
  | ForAll _ ->
      let t_ = specialize x.typ in
      unify t_ t; 
      let t = deref_type t in
      find x t
  | _ -> x
let hoge1 x t =
  match x with
  | Var x -> Var(hoge x t)
  | _ -> x

(* 使用する箇所を置き換えたり LetRec を追加したり *)
let rec g self_t e =
  match e with
  | V x -> V(hoge1 x self_t)
  | Op(x, ys) ->
      let ts = List.map typeof ys in
      let x_ = hoge1 x (Tfun(ts,self_t)) in
      Op(x_, List.map2 hoge1 ys ts)
  | Tuple xs ->
      let ts = List.map (fun _ -> new_type ()) xs in
      unify (Ttuple ts) self_t;
      Tuple(List.map2 hoge1 xs ts)
  | ExtApp(x, ys) ->
      let ts = List.map typeof ys in
      let _ = hoge x (Tfun(ts,self_t)) in
      ExtApp(x, List.map2 hoge1 ys ts)
  | If(c,x,y,e1,e2) -> If(c,x,y, g self_t e1, g self_t e2)
  | Let([],e1,e2) -> Let([], g Tunit e1, g self_t e2)
  | Let([x],e1,e2) ->
      begin match x.typ with
      | ForAll(_, t) ->
          let e2_ = g self_t e2 in
          M3.fold
            (fun t_ y e ->
              let tenv = make_tenv t t_ in
              let e1_ = copy M.empty tenv e1 in
              Let([y], g t_ e1_, e))
            (try M.find x !env with Not_found -> M3.empty)
            e2_
      | _ -> Let([x], g x.typ e1, g self_t e2) end
  | Let(xs,e1,e2) ->
      let ys = List.map (fun x -> new_id x.name) xs in
      List.fold_left2
        (fun e x y -> match x.typ with ForAll(_,t) ->
          M3.fold
            (fun t_ x_ e ->
              let zs = Shared.list_substq y x_ ys in
              let tenv = make_tenv t t_ in
              let e1_ = copy M.empty tenv e1 in
              Let(zs, g (Ttuple(List.map (fun z -> z.typ) zs)) e1_, e))
            (try M.find x !env with Not_found -> M3.empty)
            e
         | _ -> e)
        (Let(xs, g (Ttuple(List.map (fun x -> specialize x.typ) xs)) e1,
            g self_t e2))
        xs ys
  | LetRec(x,ys,e1,e2) ->
      match x.typ with
      | ForAll(_, (Tfun(ts,_) as t)) ->
          let e2_ = g self_t e2 in
          M3.fold
            (fun t_ x_ e ->
              let tenv = make_tenv t t_ in
              let ys_ = List.map (copy_id tenv) ys in
              let env = M.add x x_ (M.add_list2 ys ys_ M.empty) in
              let e1_ = copy env tenv e1 in
              match t_ with Tfun(_,tr) ->
                LetRec(x_, ys_, g tr e1_, e)
              | _ -> assert false)
            (try M.find x !env with Not_found -> M3.empty)
            e2_
      | Tfun(_,t) -> LetRec(x,ys, g t e1, g self_t e2)
      | _ -> assert false

open List

let add_builtin_fundefs e =
  let add1 e x =
    let ts = match x.typ with
      | Tfun(ts,_) | ForAll(_,Tfun(ts,_)) -> ts
      | _ -> assert false
    in
    let ys = map (fun t -> new_id_t "x" t) ts in
    let vs = map (fun y -> Var y) ys in
    let body = match x.name with
      | "=" | "==" -> If(Eq, hd vs, hd (tl vs), V(Bool true), V(Bool false))
      | "<>" -> If(Eq, hd vs, hd (tl vs), V(Bool false), V(Bool true))
      | "<=" -> If(LE, hd vs, hd (tl vs), V(Bool true), V(Bool false))
      | "<" -> If(LE, hd (tl vs), hd vs, V(Bool false), V(Bool true))
      | ">=" -> If(LE, hd (tl vs), hd vs, V(Bool true), V(Bool false))
      | ">" -> If(LE, hd vs, hd (tl vs), V(Bool false), V(Bool true))
      | l -> ExtApp(special_id_t ("." ^ l) x.typ, vs)
    in
    LetRec(x, ys, body, e)
  in
  fold_left add1 e Builtin.fun_idents

(* エントリーポイント *)
let f e =
  env := M.empty;
  let ret = g Tunit (add_builtin_fundefs e) in
  env := M.empty; (* メモリ節約 :-p *)
  ret
