(* 覗き穴最適化 *)

open TreeAsm
open LinearAsm

let all_nodes = H.create 100
let funs = H.create 100

let rec used_labels s e =
  if e == end_node then s
  else let s = match e.exp with
  | JUMP x | JAL x | BE(_,_,x) | BNE(_,_,x) | BL(_,_,x) | BGE(_,_,x)
  | BEI(_,_,x) | BNEI(_,_,x) | BLI(_,_,x) | BGEI(_,_,x)
  | FBL(_,_,x) | FBGE(_,_,x) | STORE(_,_,x) | LOAD(_,_,x)
  | ADDI(_,_,x) -> S2.add x s
  | NOP | HLT | MOV _ | GET _ | PUT _ | JR _ | ADD _ | SUB _ | SLL _ | SRR _
  | AND _ | OR _ | XOR _ | XNOR _ | FADD _ | FSUB _ | FMUL _ | FINV _
  | FSQRT _ -> s
  in
  used_labels s e.cont

let rec elim_labels s e =
  if e == end_node then ()
  else begin
    if not (S2.mem e.label s) then e.label <- nolabel;
    elim_labels s e.cont
  end

let clean xs =
  let s = List.fold_left (fun s x -> used_labels s x.entry) S2.empty xs in
  List.iter
    (fun x -> elim_labels s x.entry;
      if x.fbody.label == nolabel then x.fbody.label <- x.fname ^ ".body")
    xs

let rec h e =
  if e == end_node then ()
  else begin
    if e.label != nolabel then H.add all_nodes e.label e;
    h e.cont
  end

let modified = ref false

let rec is_tail self e =
  assert (e != end_node);
  e == self.exit ||
  match e with
  | { exp = JUMP x } when H.mem all_nodes x ->
      begin match H.find all_nodes x with
      | e_ when e_ == self.exit -> true
      | { exp = JUMP y } -> e.exp <- JUMP y; is_tail self e
      | { exp = JR r } -> r = reg_ra
      | _ -> false end
  | _ -> false

let trunc_cont e =
  while e.cont.label == nolabel do
    modified := true;
    e.cont <- e.cont.cont;
  done

let rec thr_jmp x k =
  try match H.find all_nodes x with
  | { exp = JUMP y } -> modified := true; thr_jmp y k
  | _ -> k x
  with Not_found -> k x

let rec g self e =
  if e == end_node then ()
  else begin
    begin match e with
    | { exp = JAL x; cont = e_ } when is_tail self e_ -> (* 末尾呼び出し *)
        modified := true;
        trunc_cont e;
        begin try
          let func = H.find funs x in
          if func.stk = self.stk then (* スタックフレームを放置できる *)
            e.exp <- JUMP func.fbody.label
          else raise Not_found (* トリック *)
        with Not_found ->
          if self.stk = 0 then e.exp <- JUMP x
          else begin
            e.exp <- LOAD(reg_ra, reg_sp, "0");
            e.cont <- make (ADDI(reg_sp, reg_sp, string_of_int self.stk))
                (make (JUMP x) e.cont)
          end
        end

    | { exp = JUMP x } ->
        trunc_cont e;
        thr_jmp x
          (fun x -> e.exp <- JUMP x;
            try match H.find all_nodes x with
            | { exp = JR r } -> modified := true; e.exp <- JR r
            | e1 when e1 == e.cont && e.label == nolabel ->
                modified := true;
                e.label <- e1.label;
                e.exp <- e1.exp;
                e.cont <- e1.cont
            | _ -> ()
            with Not_found -> ())
    | { exp = JR x } -> trunc_cont e

    | { exp = BE(x,y,z); cont =
        { label = l; exp = JUMP w; cont = { label = z2 } as c }}
      when z = z2 && l == nolabel ->
        modified := true; e.exp <- BNE(x,y,w); e.cont <- c
    | { exp = BNE(x,y,z); cont =
        { label = l; exp = JUMP w; cont = { label = z2 } as c }}
      when z = z2 && l == nolabel ->
        modified := true; e.exp <- BE(x,y,w); e.cont <- c
    | { exp = BL(x,y,z); cont =
        { label = l; exp = JUMP w; cont = { label = z2 } as c }}
      when z = z2 && l == nolabel ->
        modified := true; e.exp <- BGE(x,y,w); e.cont <- c
    | { exp = BGE(x,y,z); cont =
        { label = l; exp = JUMP w; cont = { label = z2 } as c }}
      when z = z2 && l == nolabel ->
        modified := true; e.exp <- BL(x,y,w); e.cont <- c
    | { exp = BEI(x,y,z); cont =
        { label = l; exp = JUMP w; cont = { label = z2 } as c }}
      when z = z2 && l == nolabel ->
        modified := true; e.exp <- BNEI(x,y,w); e.cont <- c
    | { exp = BNEI(x,y,z); cont =
        { label = l; exp = JUMP w; cont = { label = z2 } as c }}
      when z = z2 && l == nolabel ->
        modified := true; e.exp <- BEI(x,y,w); e.cont <- c
    | { exp = BLI(x,y,z); cont =
        { label = l; exp = JUMP w; cont = { label = z2 } as c }}
      when z = z2 && l == nolabel ->
        modified := true; e.exp <- BGEI(x,y,w); e.cont <- c
    | { exp = BGEI(x,y,z); cont =
        { label = l; exp = JUMP w; cont = { label = z2 } as c }}
      when z = z2 && l == nolabel ->
        modified := true; e.exp <- BLI(x,y,w); e.cont <- c
    | { exp = FBL(x,y,z); cont =
        { label = l; exp = JUMP w; cont = { label = z2 } as c }}
      when z = z2 && l == nolabel ->
        modified := true; e.exp <- FBGE(x,y,w); e.cont <- c
    | { exp = FBGE(x,y,z); cont =
        { label = l; exp = JUMP w; cont = { label = z2 } as c }}
      when z = z2 && l == nolabel ->
        modified := true; e.exp <- FBL(x,y,w); e.cont <- c

    | { exp = JAL x } -> thr_jmp x (fun x -> e.exp <- JAL x)
    | { exp = BE(x,y,z) } -> thr_jmp z (fun z -> e.exp <- BE(x,y,z))
    | { exp = BNE(x,y,z) } -> thr_jmp z (fun z -> e.exp <- BNE(x,y,z))
    | { exp = BL(x,y,z) } -> thr_jmp z (fun z -> e.exp <- BL(x,y,z))
    | { exp = BGE(x,y,z) } -> thr_jmp z (fun z -> e.exp <- BGE(x,y,z))
    | { exp = BEI(x,y,z) } -> thr_jmp z (fun z -> e.exp <- BEI(x,y,z))
    | { exp = BNEI(x,y,z) } -> thr_jmp z (fun z -> e.exp <- BNEI(x,y,z))
    | { exp = BLI(x,y,z) } -> thr_jmp z (fun z -> e.exp <- BLI(x,y,z))
    | { exp = BGEI(x,y,z) } -> thr_jmp z (fun z -> e.exp <- BGEI(x,y,z))
    | { exp = FBL(x,y,z) } -> thr_jmp z (fun z -> e.exp <- FBL(x,y,z))
    | { exp = FBGE(x,y,z) } -> thr_jmp z (fun z -> e.exp <- FBGE(x,y,z))

    | _ -> ()
    end;
    g self e.cont
  end

let f xs =
  Format.eprintf "peephole optimization...@.";
  modified := true;
  while !modified do
    modified := false;
    clean xs;
    H.clear funs;
    H.clear all_nodes;
    List.iter (fun x -> H.add funs x.fname x; h x.entry) xs;
    List.iter (fun x -> g x x.entry) xs;
  done;
  H.clear funs;
  H.clear all_nodes;
  xs
